from .__about__ import __version__
from .geo import Geo
from .geo_manager import GeoManager
from .utils import current_geo


__all__ = [
    "__version__",
    "GeoManager",
    "current_geo",
    "Geo",
]



import os
import sys
import unittest

from flask import Flask
from flask import render_template
from flask_geolocation import current_geo
from flask_geolocation import GeoManager
from flask_geolocation.__about__ import __author__
from flask_geolocation.__about__ import __author_email__
from flask_geolocation.__about__ import __copyright__
from flask_geolocation.__about__ import __description__
from flask_geolocation.__about__ import __license__
from flask_geolocation.__about__ import __maintainer__
from flask_geolocation.__about__ import __title__
from flask_geolocation.__about__ import __url__
from flask_geolocation.__about__ import __version__
from flask_geolocation.__about__ import __version_info__
from semantic_version import Version

sys_version = Version(
    major=sys.version_info.major,
    minor=sys.version_info.minor,
    patch=sys.version_info.micro,
)


class AboutTestCase(unittest.TestCase):
    """make sure we can get version and other info."""

    def test_have_about_data(self):
        self.assertTrue(__title__ is not None)
        self.assertTrue(__description__ is not None)
        self.assertTrue(__url__ is not None)
        self.assertTrue(__version_info__ is not None)
        self.assertTrue(__version__ is not None)
        self.assertTrue(__author__ is not None)
        self.assertTrue(__author_email__ is not None)
        self.assertTrue(__maintainer__ is not None)
        self.assertTrue(__license__ is not None)
        self.assertTrue(__copyright__ is not None)


class InitializationTestCase(unittest.TestCase):
    def setUp(self):
        self.app = Flask(__name__)
        self.app.config["SECRET_KEY"] = "1234"

    def test_init_app(self):
        geo_manager = GeoManager("wrong_path")
        self.assertIsInstance(geo_manager, GeoManager)

    def test_init_with_ip_data_path(self):
        geo_manager = GeoManager("tests/data")
        geo_manager.init_app(self.app, add_context_processor=True)
        self.assertIsInstance(geo_manager, GeoManager)


class NoCallbackTestCase(unittest.TestCase):
    def setUp(self):
        template_abs = os.path.abspath("tests/templates")
        self.app = Flask(__name__, template_folder=template_abs)
        self.app.config["SECRET_KEY"] = "1234"

        self.environ_base = {"REMOTE_ADDR": "154.204.60.219"}

        self.geo_manager = GeoManager("tests/data")

        self.geo_manager.init_app(self.app)

        @self.app.route("/")
        def index():
            return (
                current_geo.country_symbol
                + ", "
                + current_geo.country_name
                + ", "
                + current_geo.timezone
            )

        @self.app.route("/template")
        def template():
            return render_template("country.html")

    def test_global_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/", environ_base=self.environ_base)
            expect = "HK, Hong Kong, UTC+8"
            self.assertEqual(expect, result.data.decode("utf-8"))

    def test_template_context_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/template", environ_base=self.environ_base)
            expect = "HK, Hong Kong, UTC+8"
            self.assertEqual(expect, result.data.decode("utf-8"))


class IpCallbackTestCase(unittest.TestCase):
    def setUp(self):
        template_abs = os.path.abspath("tests/templates")
        self.app = Flask(__name__, template_folder=template_abs)
        self.app.config["SECRET_KEY"] = "1234"

        self.environ_base = {"REMOTE_ADDR": "154.204.60.219"}

        self.geo_manager = GeoManager("tests/data")
        self.geo_manager.init_app(self.app)

        @self.geo_manager.use_ip
        def ip_callback():
            """This is a ip address in Singapore"""
            return "43.156.97.49"

        @self.app.route("/")
        def index():
            return (
                current_geo.country_symbol
                + ", "
                + current_geo.country_name
                + ", "
                + current_geo.timezone
            )

        @self.app.route("/template")
        def template():
            return render_template("country.html")

    def test_global_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/", environ_base=self.environ_base)
            expect = "SG, Singapore, UTC+8"
            self.assertEqual(expect, result.data.decode("utf-8"))

    def test_template_context_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/template", environ_base=self.environ_base)
            expect = "SG, Singapore, UTC+8"
            self.assertEqual(expect, result.data.decode("utf-8"))


class TimezoneCallbackTestCase(unittest.TestCase):
    def setUp(self):
        template_abs = os.path.abspath("tests/templates")
        self.app = Flask(__name__, template_folder=template_abs)
        self.app.config["SECRET_KEY"] = "1234"

        self.environ_base = {"REMOTE_ADDR": "154.204.60.219"}

        self.geo_manager = GeoManager("tests/data")
        self.geo_manager.init_app(self.app)

        @self.geo_manager.use_timezone
        def timezone_callback():
            """change Hong Kong timezone to UTC-4"""
            return "UTC-4"

        @self.app.route("/")
        def index():
            return (
                current_geo.country_symbol
                + ", "
                + current_geo.country_name
                + ", "
                + current_geo.timezone
            )

        @self.app.route("/template")
        def template():
            return render_template("country.html")

    def test_global_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/", environ_base=self.environ_base)
            expect = "HK, Hong Kong, UTC-4"
            self.assertEqual(expect, result.data.decode("utf-8"))

    def test_template_context_with_request(self):
        with self.app.test_client() as c:
            result = c.get("/template", environ_base=self.environ_base)
            expect = "HK, Hong Kong, UTC-4"
            self.assertEqual(expect, result.data.decode("utf-8"))



from flask import g
from flask import request

from . import ip_database
from .geo import Geo
from .timezone import timezone_map
from .utils import _geo_context_processor


class GeoManager:
    def __init__(self, ip_data_path=None):
        self.ip_data_path = ip_data_path
        if self.ip_data_path is None:
            self.ip_data_path = "maxmind/data"

        self.ip_data_source = "maxmind"
        self.ip_db = None

        self._ip_callback = None
        self._timezone_callback = None

    def init_app(self, app, add_context_processor=True):
        if self.ip_db is None:
            self.ip_db = ip_database.from_source(self.ip_data_source)
            self.ip_db.load(self.ip_data_path)

        app.geo_manager = self
        if add_context_processor:
            app.context_processor(_geo_context_processor)

    def use_ip(self, callback):
        """get ip for current user, all other infos rely on ip"""
        self._ip_callback = callback
        return self._ip_callback

    @property
    def ip_callback(self):
        return self._ip_callback

    def use_timezone(self, callback):
        self._timezone_callback = callback
        return self.timezone_callback

    @property
    def timezone_callback(self):
        return self._timezone_callback

    def _load_geo(self):
        """construct geo info from ip address"""
        ip = None
        if self._ip_callback is not None:
            ip = self._ip_callback()
        else:
            if request.headers.getlist("X-Forwarded-For"):
                ip = request.headers.getlist("X-Forwarded-For")[0]
            else:
                ip = request.remote_addr

        (country_symbol, country_name) = self.ip_db.get(ip)

        timezone = timezone_map[country_symbol]
        if self._timezone_callback:
            tz = self._timezone_callback()
            if tz is not None:
                timezone = tz

        geo = Geo(
            ip=ip,
            country_symbol=country_symbol,
            country_name=country_name,
            timezone=timezone,
        )
        g._loaded_geo = geo



from flask import current_app
from flask import g
from flask import has_request_context
from werkzeug.local import LocalProxy

current_geo = LocalProxy(lambda: _get_geo())


def _get_geo():
    if has_request_context():
        if "_loaded_geo" not in g:
            current_app.geo_manager._load_geo()
        return g._loaded_geo
    return None


def _geo_context_processor():
    return dict(current_geo=_get_geo())



import unittest

from flask_geolocation import ip_database


class TestMaxmindDatabase(unittest.TestCase):
    def setUp(self):
        self.db = ip_database.from_source("maxmind")
        self.db.load("tests/data")

    def test_china(self):
        info = self.db.get("121.34.22.206")
        self.assertEqual(info, ("CN", "China"))

    def test_sigapore(self):
        info = self.db.get("43.156.97.49")
        self.assertEqual(info, ("SG", "Singapore"))

    def test_hongkong(self):
        info = self.db.get("154.204.60.219")
        self.assertEqual(info, ("HK", "Hong Kong"))



__title__ = "Flask-Geo"
__description__ = "User Geolocation for Flask"
__url__ = "https://github.com/wankai/flask-geo"
__version_info__ = ("0", "1", "0")
__version__ = ".".join(__version_info__)
__author__ = "Wankai Zhang"
__author_email__ = "wankaizhang@gmail.com"
__maintainer__ = "Wankai Zhang"
__license__ = "MIT"
__copyright__ = "(c) 2022 by Wankai Zhang"



import os
from ipaddress import ip_address


def from_source(source: str):
    if source == "maxmind":
        return MaxmindDatabase()
    raise Exception("ip source not supported")


"""
class IPDatabase:
    def __init__(self):
        return None

    def load(self, path):
        return None

    return (country_code, country_symbol, country_name)
    all three elements in tuple is None if can't find
    country info for this IP

    def get(self, ip: str):
        return None
"""


class MaxmindDatabase:
    def __init__(self):
        self._addresses = {}
        self._codes = {}

    def load(self, path):
        self._path = path

        data_dir = _find_latest_database(self._path)

        code_file = _get_code_file(data_dir)
        self._load_country_code(code_file)

        ip_file = _get_ip_file(data_dir)
        self._load_ip_dict(ip_file)

    def _load_country_code(self, path):
        with open(path) as fp:
            while True:
                line = fp.readline()
                if not line:
                    break
                arr = line.strip().split(",")
                if len(arr) != 7:
                    continue
                if arr[0] == "geoname_id":
                    continue
                if not arr[4]:
                    continue

                self._codes[int(arr[0])] = (arr[4], arr[5].strip('"'))

    def __load_ip_items(self, path):
        items = []
        with open(path) as fp:
            while True:
                line = fp.readline()
                if not line:
                    break
                arr = line.strip().split(",")
                if len(arr) != 6:
                    continue
                if arr[0] == "network":
                    continue
                ip_arr = arr[0].split("/")
                if len(ip_arr) != 2:
                    continue
                if not arr[1]:
                    continue

                mask = int(ip_arr[1])
                address = int(ip_address(ip_arr[0]))
                code = int(arr[1])
                items.append((mask, address, code))

        return items

    def __get_grouped_items(self, items):
        grouped_items = []
        last_index = -1
        group_single = []
        for mask, address, code in items:
            if last_index > -1 and mask != items[last_index][0]:
                grouped_items.append(group_single.copy())
                group_single.clear()
            group_single.append((mask, address, code))
            last_index = last_index + 1

        if group_single:
            grouped_items.append(group_single)

        return grouped_items

    def _load_ip_dict(self, path):
        items = self.__load_ip_items(path)
        items.sort(key=lambda x: x[0])

        grouped_items = self.__get_grouped_items(items)

        for single in grouped_items:
            address_set = {}
            if single:
                mask = single[0][0]

            for _, address, code in single:
                if code in self._codes:
                    address_set[address] = self._codes[code]
            self._addresses[mask] = address_set

    def get(self, ip: str):
        ip_int = int(ip_address(ip))
        for mask_digit_num, address_map in self._addresses.items():
            mask = _mask_map[mask_digit_num]
            ip_after_mask = ip_int & mask
            if ip_after_mask in address_map:
                return address_map[ip_after_mask]


def _find_latest_database(path):
    all_dir1 = os.listdir(path)
    all_dir = map(lambda x: path + "/" + x, all_dir1)
    items = map(lambda x: (os.path.getmtime(x), x), all_dir)
    return max(items)[1]


def _get_code_file(path):
    return path + "/GeoLite2-Country-Locations-en.csv"


def _get_ip_file(path):
    return path + "/GeoLite2-Country-Blocks-IPv4.csv"


_mask_map = {
    31: 0xFFFFFFFE,
    30: 0xFFFFFFFC,
    29: 0xFFFFFFF8,
    28: 0xFFFFFFF0,
    27: 0xFFFFFFE0,
    26: 0xFFFFFFC0,
    25: 0xFFFFFF80,
    24: 0xFFFFFF00,
    23: 0xFFFFFE00,
    22: 0xFFFFFC00,
    21: 0xFFFFF800,
    20: 0xFFFFF000,
    19: 0xFFFFE000,
    18: 0xFFFFC000,
    17: 0xFFFF8000,
    16: 0xFFFF0000,
    15: 0xFFFE0000,
    14: 0xFFFC0000,
    13: 0xFFF80000,
    12: 0xFFF00000,
    11: 0xFFE00000,
    10: 0xFFC00000,
    9: 0xFF800000,
    8: 0xFF000000,
    7: 0xFE000000,
    6: 0xFC000000,
    5: 0xF8000000,
    4: 0xF0000000,
    3: 0xE0000000,
    2: 0xC0000000,
    1: 0x80000000,
    0: 0x00000000,
}



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="Flask-Geolocation",
    install_requires=[
        "Flask>=1.0.4",
        "Werkzeug>=1.0.1",
    ],
)
